#!/bin/bash


## NMERASP - nmerasp.sh: read data from the sense-hat addon card of the
# raspberry-pi and output it to NMEA standard form.
## Type /usr/local/sbin/nmerasp.sh --help (or -h) for use.

## COPYRIGHT 2016-2017 Free Software Foundation, Inc. License GNU.
## GPL version 3 or later <http://gnu.org/licenses/gpl.html>.
## This is a free software: you are free to use, change and redistribute it.
## There is NO WARRANTY, to the extent permitted by law.
## Navigation beta software, use it as under your own responsabilty !
## Frank Baudet <http://marinux.tuxfamily.org>.

## HISTORY
## 17/04/21 first public version > V0.2

########################################################################


                              ############
                              ### HELP ###
                              ############

disp_help() {

   help='
 ¤
 ¤ ######################## NMERASP V0.2 HELP ##########################
 ¤
 ¤ <nmerasp.sh > read data from the Sense-Hat addon card of the Raspberry
 ¤ PI and output it to NMEA standard form:
 ¤
 ¤    - Compass (HDT)
 ¤    - Barometer and Hygrometer (MDA)
 ¤    - Pitch and Roll (XDR).
 ¤
 ¤    Usage: < sudo /usr/local/sbin/nmerasp.sh start[&]|stop|-h|--help >
 ¤
 ¤ At the first start a default configuration file /usr/local/etc/
 ¤ nmerasp.conf is created. This file contain complementary informations.
 ¤ Edit it to adapt the output port wished, using a virtual serial port
 ¤ (default) directly usable in a chartplotter as OpenCPN or mix this
 ¤ output trought a NMEA multiplexer as Muplex or kPlex.
 ¤
 ¤ Superuser rights are required to configure and output ports.
 ¤
 ¤ Note: To avoid orphans process, aka zombies, take carefull to close
 ¤ NMErasp using < nmerasp.sh stop >, not with < CTRL+C >.
 ¤
 ¤ Depend of:
 ¤
 ¤    - RTIMULibDrive11 by Serge Schneider from librtimulib-utils package.
 ¤    - gawk package based on awk by Aho, Kernighan and Weinberger.
 ¤    - socat package by Gerhard Reiger.
 ¤
 ¤ And for the optional Graphical User Interface NMErasp-GUI:
 ¤    - yad package.
 ¤    - xdg-open from xdg-utils package.
 ¤    - gksudo from gksu package.
 ¤
 ¤ NMERASP run under any Linux distro on Raspberry Pi.
 ¤ Tested with Xinutop 2.0, based on Xfce Debian Jessie running on a
 ¤ Raspberry Pi 3 equiped with a standard Sense-Hat addon card.
 ¤
 ¤ INSTALLATION: NMErasp is provided in a tar.gz archive (tgz) including
 ¤ GUI, menu, icons, so use:
 ¤    < sudo tar -xvz -C / -f nmerasp-x.x.tgz >
 ¤ Or use a package manager as PACKMGN which is distribution independent
 ¤ and include an uninstall function.
 ¤ It may be necessary to restart your WM session to display the new
 ¤ menu entry.
 ¤ Advanced users who need NMErasp only in command line can extract the
 ¤ archive in a temp directory (ex: option -C /tmp/) and copy nmerasp.sh,
 ¤ the only one necessary file, in /usr/local/sbin.
 ¤
 ¤ GPL 3+, Navigation beta software, use it under your own responsability !
 ¤ Frank Baudet <http://marinux.tuxfamily.org>
 ¤
 ¤
 ¤ ##################### MAGNETOMETER CALIBRATION ###################### 
 ¤
 ¤ To use correctly compass informations, it`s necessary to calibrate it.
 ¤ Here is the howto, adapted from Raspberry official documentation.
 ¤
 ¤ Copy all necessary files in your home directory and go inside
 ¤    < cp /usr/share/librtimulib-utils/RTEllipsoidFit ./ -a >
 ¤    < cd RTEllipsoidFit >
 ¤ Launch the program calibration
 ¤    < RTIMULibCal >
 ¤ You will then see this menu:
 ¤
 ¤ Options are:
 ¤
 ¤ m - calibrate magnetometer with min/max
 ¤ e - calibrate magnetometer with ellipsoid (do min/max first)
 ¤ a - calibrate accelerometers
 ¤ x - exit
 ¤
 ¤ Enter option:
 ¤ Press lowercase m. The following message will then show; press any
 ¤ key to start.
 ¤
 ¤   Magnetometer min/max calibration
 ¤   --------------------------------
 ¤   Waggle the IMU chip around, ensuring that all six axes
 ¤   (+x, -x, +y, -y and +z, -z) go through their extrema.
 ¤   When all extrema have been achieved, enter `s` to save, `r` to reset
 ¤   or `x` to abort and discard the data.
 ¤
 ¤   Press any key to start...
 ¤ After it starts, you will see something similar to this scrolling up
 ¤ the screen:
 ¤
 ¤ Min x:  51.60  min y:  69.39  min z:  65.91
 ¤ Max x:  53.15  max y:  70.97  max z:  67.97
 ¤ Focus on the two lines at the very bottom of the screen, as these are
 ¤ the most recently posted measurements from the program. Now you have
 ¤ to move the Astro Pi around in every possible way you can think of.
 ¤ It helps if you unplug all non-essential cables to avoid clutter.
 ¤
 ¤ Try and get a complete circle in each of the pitch, roll and yaw axes.
 ¤ Take care not to accidentally eject the SD card while doing this.
 ¤ Spend a few minutes moving the Astro Pi, and stop when you find that
 ¤ the numbers are not changing anymore.
 ¤
 ¤ Now press lowercase s then lowercase x to exit the program. If you run
 ¤ the  ls command now, you`ll see a new RTIMULib.ini file has been created.
 ¤
 ¤ In addition to those steps, you can also do the ellipsoid fit by
 ¤ performing the steps above, but pressing e instead of m.
 ¤
 ¤ When you`re done, copy the resulting RTIMULib.ini to /usr/local/etc/
 ¤ 	  < sudo cp RTIMULib.ini /usr/local/etc >
 ¤ You are now done.
 ¤
   \n'

   printf "$help"

}

############################### END HELP ###############################


                              ############
                              ### CONF ###
                              ############

create_conf() {

   conf='

 ##################### NMERASP V0.2 CONFIGURATION ######################

 ## Two ## is a comment.
 ## One # is an option, uncomment/comment to activate/deactivate.
 ## <...> is a command or code line.
 ## Default values convenient for most of cases.
 ## At minimum, adjust your output to use directly in a chartplotter
 ## as OpenCPN or a multiplexer as Muplex or kPlex.


 #################
 ## output port ##
 #################

 ## Data output port: can be a real port as /dev/ttySx, /dev/ttyUSBx,
 ## /dev/rfcommx or virtual port as /dev/ttyPSx (which will be created),
 ## where x=port number. This virtual port can be read directly by
 ## OpenCPN or handled as input by Muplex or kPlex.
 ## Type < ls /dev/ttyS* or /dev/ttyUSB* or /dev/rfcomm* > to examine
 ## your devices.
 ## Note: Raspberry (hard) serial port is named /dev/ttyAMA0.

 ## here is a virtual port.
 out_port=/dev/ttyPS0

 ## virtual port if using NMErasp-GUI, which have it own control
 ## terminal, leave blank if unused
 #term_port=
 term_port=/dev/ttyPS10

 ## baudrate speed output (4800, 9600, 19200, 38400).
 ## 4800=standard NMEA, 38400=NMEA High Speed (HS) as AIS
 spd_out=4800


 ##################
 ## output rates ##
 ##################

 ## NMEA rates sentences, every x seconds
 ## leave the value blank if no output wished

 ## HDM (compass)
 freq_yaw=1

 ## XDR (in "A"=Angular mode, pitch and roll)
 freq_proll=2

 ## MDA (barometer and humidity)
 freq_mto=10

 ## NOTES:
 ## OpenCPN (V4.4.0) don`t display barometric pressure in text and jauge
 ## but OK in history mode.
 ## The temperature sensor is unused: it`s not a navigational datum,
 ## giving only the Pi box temperature.


 #########################
 ## initial corrections ##
 #########################

 ## Integer, may be negative.

 ## magnetic variation: east [+], west - in degrees
 var_mag=0
 ## roll in degrees
 cor_rol=0
 ## pitch in degrees
 cor_pit=0
 ## barometer in mb
 cor_baro=0
 ## humidity in percent
 cor_hum=0

 ## pitch sign: 1 or -1, depend of your sense-hat dispositon on board,
 ## nose up, nose down in OpenCPN.
 sig_pit=1
 ## roll sign: 1 or -1, depend of the dispostion of your sense hat on
 ## board (tribord ">" or babord "<" in OpenCPN)
 sig_rol=-1

 ## Note:
 ## to start automatically NMErasp, add:
 ##   < /usr/local/sbin/nmerasp.sh start & >
 ## in your /etc/rc.local.
 ## If using a multiplexer, NMErasp must be started before, or use a
 ## temporization as in Muplex.

 \n'

   # create dir(s) for conf if necessary
   [ -d "$conf_path" ] || mkdir -p "$conf_path"
   printf "$conf" > "$conf_path/$conf_file"
   # full access
   chmod 666 "$conf_path/$conf_file"
}

############################### END CONF ###############################


                            ################
                            ## START STOP ##
                            ################

# close all properly
terminator() {
# kill all child process
if [ -f /run/lock/nmerasp.pid ]; then
   # get group pid
   #grp_proc=$(ps -ejH | grep nmerasp.sh | cut -d ' ' -f2) # unusable, so
   grp_proc=$(cat </proc/"$(cat </run/lock/nmerasp.pid)"/stat |
      cut -d ' ' -f5)
   for i in socat gawk; do
      # list and kill childs process
      procs_childs="${procs_childs} $(pgrep -d ' ' -x -g ${grp_proc} $i)"
   done
   [ "$procs_childs" = "" ] || kill ${procs_childs}
   # remove tracks
   rm /run/lock/nmerasp.pid
fi
# time to register error by GUI
sleep 1
[ -f /tmp/nmerasp.error ] && rm /tmp/nmerasp.error
# only one instance for all the system
[ $(pgrep -x -c RTIMULibDrive11) -gt 0 ] &&
   { killall RTIMULibDrive11 && wait; } 2>/dev/null
# in case of many instances
killall -wq nmerasp.sh
}

# cmd start during nmerasp already and regulary running
if [ "$1" = "start" ]; then
   if [ $(pgrep -x -c nmerasp.sh) -gt 1 ] && [ -f /run/lock/nmerasp.pid ]
   then
      printf "\n   NMErasp is already running. \n\n"
      exit
   fi

# cmdline stop
elif [ "$1" = "stop" ]; then
   # nmerasp is running
   if [ $(pgrep -x -c nmerasp.sh) -gt 1 ] || [ -f /run/lock/nmerasp.pid ]
   then
      printf "\n   Stopping NMErasp \n\n"
      terminator
   # nmerasp not running
   else
      printf "\n   NMErasp is already stopped. \n\n"
      exit
   fi

# cmdline help
elif [ "$1" = "--help" ] || [ "$1" = "-h" ]; then
   disp_help | more
   exit

# error in cmdline with exception to start, already handled
elif [ "$1" != "start" ]; then
   printf "\n   Error ! in command line
      usage: < nmerasp.sh start[&]|stop|-h|--help >. \n\n"
   exit
fi

# cleanup, in case
[ $(pgrep -x -c RTIMULibDrive11) -gt 0 ] && killall -w RTIMULibDrive11
[ -f /tmp/nmerasp.error ] && rm /tmp/nmerasp.error
# in case, for others distros, same attributes as Debian
chmod 1777 /run/lock
# create current nmerasp.sh pid file
echo $$ >/run/lock/nmerasp.pid

############################ END START STOP ############################


                           ###################
                           ### PRELIMINARY ###
                           ###################

# need to be run as superuser
if [ $(id -u) -ne 0 ]; then
   printf "\n   Error ! NMERASP require superuser rights,
      use < sudo /usr/local/sbin/nmerasp.sh >. \n\n"
   terminator
fi

# look for depends
for i in RTIMULibDrive11 socat gawk; do
   if ! (which "$i" >/dev/null) ; then
      j="$j $i"
   fi
done
if [ -n "$j" ]; then
   printf "\n missing $j ! \n\n" | tee /tmp/nmerasp.error
   terminator
fi

########################### END PRELIMINARY ############################


                            ################
                            ### SETTINGS ###
                            ################

#clear
LC_ALL=C
# debug
#set -x

# config path
conf_path=/usr/local/etc
# config file
conf_file=nmerasp.conf

# default conf creation at first start
[ -f "$conf_path/$conf_file" ] || create_conf

# load variables from conf
eval $(grep -v \# "$conf_path/$conf_file")

# standard serial NMEA input parameters
params_port="raw,clocal=1,cs8,cstopb=0,parenb=0,crtscts=0,icanon=1,echo=0"

############################# END SETTINGS #############################


                            ###############
                            ### CHECKUP ###
                            ###############

# verification baudrate
error_bdrate=1
for j in 4800 9600 19200 38400; do
   if [ ${spd_out} -eq $j >/dev/null 2>&1 ]; then
      error_bdrate=0
      break
   fi
done
if [ ${error_bdrate} -eq 1 ]; then
   printf "\n  Error ! ${spd_out}: incorrect port baudrate \n\n" |
      tee /tmp/nmerasp.error
   terminator
fi

# start RTIMULibDrive11 in the location of RTIMULib.ini
cd "$conf_path"

#test presence Raspi sense-hat TODO test in real
RTIMULibDrive11 >/dev/null 2&>1 &
if [ $? -ne 0 ]; then
   printf "\n  Error ! No Sense Hat found. \n\n" |
      tee /tmp/nmerasp.error
   terminator
else
    { killall RTIMULibDrive11 && wait; } 2>/dev/null
fi

############################# END CHECKUP ##############################


                              ############
                              ### INIT ###
                              ############

printf "\n   Starting NMERasp \n\n"

# final command assembly w/o control terminal for nmerasp-gui
if [ -n "$term_port" ]; then
   final_send="tee >(socat -u -d -d - PTY,link=$out_port,b${spd_out},${params_port},group=dialout,mode=666) |
      socat -u -d -d - PTY,link=$term_port,b${spd_out},${params_port},group=dialout,mode=666"
else
   final_send="socat -u -d -d - PTY,link=$out_port,b${spd_out},${params_port},group=dialout,mode=666"
fi

############################### END INIT ###############################


                     ###############################
                     ### GET DATA FROM SENSE HAT ###
                     ###############################

RTIMULibDrive11 | stdbuf -o0 gawk -F "," -v out_port=${out_port} \
                              -v term_port=${term_port} \
                              -v freq_yaw=${freq_yaw} \
                              -v var_mag=${var_mag} \
                              -v freq_proll=${freq_proll} \
                              -v cor_rol=${cor_rol} \
                              -v cor_pit=${cor_pit} \
                              -v sig_pit=${sig_pit} \
                              -v sig_rol=${sig_rol} \
                              -v freq_mto=${freq_mto} \
                              -v cor_baro=${cor_baro} \
                              -v cor_hum=${cor_hum} \
' BEGIN {
   # flag updated for nmerasp-gui
   system("sleep 1; touch /run/lock/nmerasp.pid")
   # create ascii board for checksum
   for(i = 0; i < 128; i++) tab_asc[sprintf("%c",i)] = i
   # initialize refresh values as integer
   timer = 0
   tim_yaw = 0
   tim_proll = 0
   tim_mto = 0
}

######################### CHECKSUM CALCULATION #########################


function xor_sent(nmea_sent) {
   # extract sentence char by char
   for ( i=1; i <= length(nmea_sent); i++ ) {
      # convert to the ASCII value from previous board
      val_asc = tab_asc[(substr(nmea_sent,i,1))]
      # memorize only the first value
      if (i == 1) check_dec = val_asc
      # xor char by (previous result) char
      else check_dec = xor(check_dec,val_asc)
   }
   # decimal to hexadecimal conversion
   check_hex=sprintf("%X", check_dec)
   # send result
   return check_hex
}

################################# YAW ##################################


$1 ~ /Sample/ {

   # refresh time drained
   if ((freq_yaw != "") && ((timer - tim_yaw) >= freq_yaw)) {
      # catch yaw value
      pos_sepy = index($3,"yaw:")
      if (pos_sepy > 0) val_yaw = substr($3,(pos_sepy + 4),length($3))
      # initial correction
      val_yaw = (int(val_yaw)) + var_mag
      # correct negatives values to 360°
      if (val_yaw < 0) val_yaw = 360 + val_yaw
      #print "yaw " val_yaw

      # HDM sentence composition
      hdm = "IIHDT" "," val_yaw ".0,T"
      # checksum calculation
      xor_sent(hdm)

      # output final HDM sentence
      hdm_sent = "\r$" hdm "*" check_hex
      print hdm_sent

      # reinit timer
      tim_yaw = systime()
   }

############################# ROLL & PITCH #############################


   # refresh time drained
   if ((freq_proll != "") && ((timer - tim_proll) >= freq_proll)) {
      # catch roll
      pos_sepr = index($1,"roll:")
      if (pos_sepr > 0) val_roll = substr($1,(pos_sepr + 5),length($1))
      # initial correction
      val_roll = ((int(val_roll)) + cor_rol) * (sig_rol)
      #print "roll " val_roll

      # catch pitch
      pos_sept = index($2,"pitch:")
      if (pos_sept > 0) val_pitch = substr($2,(pos_sept + 6),length($2))
      # initial correction
      val_pitch = ((int(val_pitch)) + cor_pit) * (sig_pit)
      #print "pitch " val_pitch

      # XDR sentence composition (A=AngularDisplacementTransducer)
      proll = "IIXDR,A," val_pitch ",,PTCH,A," val_roll ",,ROLL"
      # checksum calculation
      xor_sent(proll)

      # output final XDR sentence
      xdr_sent = "\r$" proll "*" check_hex
      print xdr_sent

      # reinit timer
      tim_proll = systime()
   }
}

######################## PRESSURE & HUMIDITY ###########################


$1 ~ /Pressure/ {

   # refresh time drained
   if ((freq_mto != "") && ((timer - tim_mto) >= freq_mto)) {
      # catch pressure
      pos_sepp = index($1,"Pressure:")
      if (pos_sepp > 0) val_pres = substr($1,(pos_sepp + 9),length($1))
      # initial correction
      val_pres = val_pres + cor_baro
      # convert mb to b
      val_pres = (val_pres / 1000 )
      #print val_pres "mb"

      # catch humidity
      val_hum = $4
      pos_seph = index($4,"humidity:")
      if (pos_seph > 0) val_hum = substr($4,(pos_seph + 9),length($4))
      # initial correction
      val_hum = (int(val_hum)) + cor_hum
      #print val_hum "%"

      # temperature
      # unused, give only the pibox temp

      # MDA sentence composition
      mda = "IIMDA,,I," val_pres ",B,,C,,C," val_hum ",,,C,,T,,M,,N,,M"
      # checksum calculation
      xor_sent(mda)

      # output final MDA sentence
      mda_sent = "\r$" mda "*" check_hex
      print mda_sent

      # reinit timer
      tim_mto = systime()
   }
}

######################### COUNTER & FINAL SEND #########################

{
  timer = systime()
}' | eval "$final_send"

exit $?
